/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* NOTE: printprimdata and printpriminfo could be merged, dont'
 * know if it will speed up the thing, maybe if they get inlined
 * they are faster separated (no "what I should print" guessing)
 * but if they are merged and they are inlined then the file will
 * be both bigger and with 2 extra guessing. Haaa undestanding
 * today compiler optimizations is hard
 */

#include "BDFile.h"

double Primitive::getdata(int x, int y) {
  double localdata;
  localdata = this->data[y+3*x];
  return localdata;
}

void Primitive::setdata(int x, int y, double value) {
  this->data[y+3*x] = value;
}

int Primitive::allocatedata(int ndata) {
  if(this->data == NULL) {
    if(this->maxdata == 0) {
      fprintf(stdout, "Allocating a new mem chunk for [i|s|h]data for " \
                      "%i elements (%i bytes): ", ndata, ndata*3*sizeof(double));
      this->data = (double *)calloc(sizeof(double), ndata * 3);
      if(this->data == NULL) { 
	fprintf(stdout, "Failed\n");
	fprintf(stderr, "[i|s|h]data allocation failed\n");
	return -1;  // memory allocation failed
      }
      else {
	fprintf(stdout, "Success\n");
	this->maxdata = ndata;
	return 0; // OK!
      }
    }
    else {
      fprintf(stderr, "WTF?! data is null but maxdata is not 0!!\n");
      return -2; // Primitive data mismatch
    }
  }
  else {
    if(ndata < (this->maxdata) * 2) { 
      /* we dont' reallocate memory if the mem avaiable is allready 
       * big enougth but less than double of avaiable. We waste but 
       * not too much
       */ 
      fprintf(stdout, "data chunk was allready enougth big (but less" \
                      "than double), we waste mem but save cpu\n");
      return 0;
    }
    else {
      fprintf(stdout, "Reallocating old mem chunk of %i elements (%i bytes)" \
                      "to a new chunk of %i elements (%i bytes): ", \
                       this->maxdata, this->maxdata*3*sizeof(double),\
                       ndata, ndata*3*sizeof(double));
      this->data = (double *)realloc(this->data, 3*sizeof(double)*ndata);
      if(this->data == NULL) {
	fprintf(stdout, "Failed\n");
	fprintf(stderr, "[i|s|h]data reallocation failed\n");
	return -1;  // memory reallocation failed
      }
      else {
	fprintf(stdout, "Success\n");
	this->maxdata = ndata;
	return 0; // OK!
      }
    }
  }
  return -3; // How the hell we got here???!!!!
}

BDFile::BDFile() {
  bfilename[0] = '\0';
  dfilename[0] = '\0';
  firstprimitive = NULL;
  lastprimitive = NULL;
  vertexs = 0;
  elements = 0;
  globalprimid = 1;
}

/* PRIMITIVE MANGEMENT */

Primitive *BDFile::new_prim(int type) {
  Primitive *prim = NULL;
  elements++;
  vertexs++;
  switch(type) {
  case PR_LINE:
    prim = new Line;
    break;
  case PR_CIRCLE:
    prim = new Circle;
    break;
  case PR_ELLIPSE:
    prim = new Ellipse;
    break;
  case PR_DATA:
    prim = new Data;
    break;
  default:
    elements--;
    vertexs--;
    fprintf(stderr, "Unknown primitive type\n");
  }
  return prim;
}

Primitive *BDFile::add_prim(int type) {
  Primitive *newprim;
  newprim = new_prim(type);
    if(firstprimitive == NULL){
      firstprimitive = newprim;
    }
    else {
      lastprimitive->nextprimitive = newprim;
    }
    newprim->prevprimitive = lastprimitive;
    lastprimitive = newprim;
    newprim->id = globalprimid++;
    return newprim;
}

void BDFile::del_prim(Primitive *oldprim) {
  if(oldprim->nextprimitive != NULL) {
    oldprim->prevprimitive->nextprimitive = oldprim->nextprimitive;
  }
  else if (oldprim->nextprimitive == NULL) {
  oldprim->prevprimitive->nextprimitive = NULL;
  }
  delete oldprim;
}


int BDFile::openbfile() {
  bfile = fopen(bfilename, "w+");
  if(bfile == NULL) {
    fprintf(stderr, "Can't open boundary file\n");
    return -1;
  }
  return 0;
}

int BDFile::opendfile() {
  dfile = fopen(dfilename, "w+");
  if(dfile == NULL) {
    fprintf(stderr, "Can't open domain file\n");
    return -1;
  }
  return 0;
}

/* FILE GENERATION */

inline void BDFile::printhash(FILE *output) {
  fprintf(output, "#\n");
}

inline void BDFile::printgeom(FILE *output) {
  fprintf(output, "#\n" \
                  "######## GEOMETRY###########\n" \
	          "#\n"\
	  );
}

inline void BDFile::printtopo(FILE *output) {
  fprintf(output, "#\n" \
                  "######## TOPOLOGY AND METRIC#######\n" \
	          "#\n"\
	  );
}

inline void BDFile::printnvne(FILE *output) {
  fprintf(output, "nv      ne\n");
}

void BDFile::printverelem(FILE *output) {
  if(vertexs > 9) fprintf(output, "%i", vertexs);
  else fprintf(output, " %i", vertexs);
  if(elements > 9) fprintf(output, "      %i\n", elements);
  else fprintf(output, "      %i\n", elements);
}

inline void BDFile::printish(FILE *output) {
  fprintf(output, "   idata   sdata   hdata\n");
}

inline void BDFile::printenbvev(FILE *output) {
  fprintf(output, "       e       n      bv      ev\n");
}

void BDFile::printprimg(FILE *output, Primitive *curprim) {
  printbegin(output, curprim->id);
  switch(curprim->type) {
    case PR_LINE:
      fprintf(output, "line\n");
      fprintf(output, "%i %.2f %.2f %.2f %.2f\n", ((Line *)curprim)->spline, ((Line *)curprim)->x1, ((Line *)curprim)->y1, ((Line *)curprim)->x2, ((Line *)curprim)->y2);
      printend(output);
      break;
    case PR_CIRCLE:
      fprintf(output, "circle\n");
      fprintf(output, "%i %.2f %.2f %.2f %i %i\n", ((Circle *)curprim)->spline, ((Circle *)curprim)->centerx, ((Circle *)curprim)->centery, ((Circle *)curprim)->radius, ((Circle *)curprim)->startarc, ((Circle *)curprim)->endarc);
      printend(output);
      break;
    case PR_ELLIPSE:
      fprintf(output, "ellipse\n");
      fprintf(output, "%i %.2f %.2f %.2f %.2f %i %i\n", ((Ellipse *)curprim)->spline, ((Ellipse *)curprim)->centerx, ((Ellipse *)curprim)->centery, ((Ellipse *)curprim)->semix, ((Ellipse *)curprim)->semiy, ((Ellipse *)curprim)->startarc, ((Ellipse *)curprim)->endarc);
      printend(output);
      break;
    case PR_DATA:
      fprintf(output, "data\n");
      fprintf(output, "NO DATA!\n"); // still to write
      printend(output);
      break;
    default:
      fprintf(stderr, "PRINTING: Unknown primitive type\n");
    }
}

void BDFile::printprimd(FILE *output, Primitive *curprim) {
  printbegin(output, curprim->id);
  printenbvev(output);
  fprintf(output, "       %i       %i       *       *\n", curprim->id, curprim->curdata); // endvertex beginvertex missing!
  printish(output);
  if(curprim->data != NULL) {
    for(int j = 0; j < curprim->curdata; j++) {
      fprintf(output, "       %.0f    %.2f    %.2f\n", curprim->getdata(j, 0), curprim->getdata(j, 1), curprim->getdata(j, 2));
    }
  }
  printend(output);
}

void BDFile::printbegin(FILE *output, int id) {
  fprintf(output, "BEGIN %i\n", id);
}

inline void BDFile::printend(FILE *output) {
  fprintf(output, "END\n");
}

void BDFile::printpriminfo(FILE *output) {
  Primitive *currentprimitive;
  currentprimitive = firstprimitive;
  for(int i = 1; i <= elements; i++) {
    printprimg(output, currentprimitive);
    if(currentprimitive->nextprimitive != NULL) {
      printhash(output);
      currentprimitive = currentprimitive->nextprimitive;
    }
    else if (currentprimitive->nextprimitive == NULL && i != elements) {
      fprintf(stderr, "printpriminfo: WTF!? How we are here?, elements say we still have" \
                      "primitive to print but currentprimitive->nextprimitive" \
                      "pointer is NULL\n");
    }
  }
}

void BDFile::printprimdata(FILE *output) {
  Primitive *currentprimitive;
  currentprimitive = firstprimitive;
  for(int i = 1; i <= elements; i++) {
    printprimd(output, currentprimitive);
    if(currentprimitive->nextprimitive != NULL) {
      printhash(output);
      currentprimitive = currentprimitive->nextprimitive;
    }
    else if (currentprimitive->nextprimitive == NULL && i != elements) {
      fprintf(stderr, "printprimdata: WTF!? How we are here?, elements say we still" \
                      "have primitive to print but currentprimitive->nextprimitive" \
                      "pointer is NULL\n");
    }
  }
}

/* FILE OUTPUT */

void BDFile::addboundary(char *buffer) {
  char *temp;
  char *filename;
  //  char *path;
  int pathlen;
  int bufferlen;
  int filelen;
  bufferlen = strlen(buffer);
  if(bufferlen >= 246) {  //boundary. is 9 letters
    fprintf(stderr, "Filename too long: it will be truncated\n");
    bufferlen = 246;
  }
  temp = rindex(buffer, '/');   // add a define for portability wrindows use teh wrong slash '\'
  if(temp != NULL) {
    filename = temp +1;
  }
  else {
    filename = buffer;
  }
    //    path = buffer;
  filelen = strlen(filename);
  temp = (char *)malloc(filelen + 1);
  pathlen = bufferlen - filelen;
  strcpy(temp, filename);
  strncpy(filename, "boundary.", 9);
  strncpy(filename + 9, temp, 245 - pathlen); //245 need space for \0
  //  filename[filelen + 10] = '\0'; look unuseful, look like allready terminated correctly
  fprintf(stdout, "Filename modified to: %s\n", buffer);
  free(temp);
}


int BDFile::outputbfile() {
  /* Check for a valid filename */
  if(bfilename[0] == '\0') {
    fprintf(stderr, "outputbfile: Invalid filename\n");
    return -1; // invalid filename
  }
  if(strstr(bfilename, "boundary.") == NULL) {
    fprintf(stdout, "outputbfile: filename does not contain \"boundary.\", adding it\n");
    addboundary(bfilename);
  }
  /* Open the file */
  if(openbfile() != 0) {
    fprintf(stderr, "error opening the boundary file\n");
    return -2;  // file opening failed
  }
  /* Print the file */
  printnvne(bfile);
  printverelem(bfile);
  printgeom(bfile);
  printpriminfo(bfile);
  printtopo(bfile);
  printprimdata(bfile); //write this
  return 0;
}

